/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : main_PSP.c
Purpose : Example application for Stack Overflow Prevention

Literature:
  [1]  https://wiki.segger.com/Stack_Overflow_Prevention

Additional information:
  This application demonstrates prevented stack overflows 
  in the main application running on the process stack 
  and in an interrupt running on the main stack.
  
  On process stack overflow the main application is kept 
  in the endless of the error handler.
  Interrupts can still be serviced, when they are re-enabled 
  in the error callback.
  
  On main stack overflow in the interrupt handler
  the system is completely halted and should be reset.

  Expected output:
    MAIN: ac has 0 of 0 bits set.
    MAIN: ac has 0 of 1 bits set.
    MAIN: ac has 1 of 2 bits set.
    MAIN: ac has 1 of 3 bits set.
    MAIN: ac has 2 of 4 bits set.
    SYSTICK: ac has 0 of 0 bits set.
    MAIN: ac has 2 of 5 bits set.
    MAIN: ac has 3 of 6 bits set.
    MAIN: ac has 3 of 7 bits set.
    MAIN: ac has 4 of 8 bits set.
    SYSTICK: ac has 0 of 1 bits set.
    MAIN: ac has 5 of 9 bits set.
    MAIN: ac has 5 of 10 bits set.
    MAIN: ac has 6 of 11 bits set.
    SYSTICK: ac has 1 of 2 bits se
    MAIN: ac has 7 of 13 bits set.
    MAIN: ac has 7 of 14 bits set.
    MAIN: ac has 8 of 15 bits set.
    SYSTICK: ac has 1 of 3 bits set.
    MAIN: ac has 8 of 16 bits set.
    MAIN: ac has 8 of 17 bits set.
    MAIN: ac has 9 of 18 bits set.
    SYSTICK: ac has 2 of 4 bits set.
    MAIN: ac has 9 of 19 bits set.
    MAIN: ac has 10 of 20 bits set.
    MAIN: ac has 10 of 21 bits set.
    SYSTICK: ac has 2 of 5 bits set.
    SYSTEM ABORT: Stack Overflow Prevented at 0x08000948.
                  SP:      0x2001F838
                  Limit:   0x2001F858
    SYSTICK: ac has 3 of 6 bits set.
    SYSTICK: ac has 3 of 7 bits set.
    SYSTICK: ac has 4 of 8 bits set.
    SYSTICK: ac has 5 of 9 bits set.
    SYSTICK: ac has 5 of 10 bits set.
    SYSTICK: ac has 6 of 11 bits set.
    SYSTICK: ac has 6 of 12 bits set.
    SYSTICK: ac has 7 of 13 bits set.
    SYSTICK: ac has 7 of 14 bits set.
    SYSTICK: ac has 8 of 15 bits set.
    SYSTICK: ac has 8 of 16 bits set.
    SYSTICK: ac has 8 of 17 bits set.
    SYSTICK: ac has 9 of 18 bits set.
    SYSTICK: ac has 9 of 19 bits set.
    SYSTICK: ac has 10 of 20 bits set.
    SYSTICK: ac has 10 of 21 bits set.
    SYSTEM ABORT: Stack Overflow Prevented at 0x08000948.
                  SP:      0x2001FC38
                  Limit:   0x2001FC58
*/

#include <stdio.h>
#include <stdlib.h>

#include "SEGGER_STOP.h"

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#define SYSTICK_CTRL (*(volatile unsigned int*)0xE000E010)
#define SYSTICK_LOAD (*(volatile unsigned int*)0xE000E014)
#define SYSTICK_VAL  (*(volatile unsigned int*)0xE000E018)

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/
static unsigned int _Calc(unsigned char* p, int Pos) {
  int Index;
  int Bit;
  int Result;
  
  if (Pos < 0) {
    return 0;
  }
  
  Index = Pos / 8;
  Bit = Pos % 8;
  Result = p[Index] & (1 << Bit) ? 1 : 0;
  
  return Result + _Calc(p, Pos -1);
}

static int _GetBits(const char* sContext, int NumBits) {
  int r;
  unsigned char ac[] = {0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55};
  
  r = _Calc(ac, NumBits - 1);
  printf("%s: ac has %d of %d bits set.\n", sContext, r, NumBits);
  
  return r;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*   ISR for the SysTick.
*   
*/
void SysTick_Handler(void);  
void SysTick_Handler(void) {
  static int _Ticks;
  
  _GetBits("SYSTICK", _Ticks++ % 80);
}

/*********************************************************************
*
*       main_app()
*
*  Function description
*   Application entry point.
*   
*/
void main_app(void);
void main_app(void) {
  int i;
#if !defined (__SEGGER_LINKER)
  //
  // Initialize stack limits if not done by runtime init.
  // Required to set PSP limit.
  //
  __SEGGER_STOP_X_InitLimits();
#endif
  //
  // Enable SysTick timer interrupt
  //
  SYSTICK_LOAD  = 10000;
  SYSTICK_VAL   = 0;
  SYSTICK_CTRL  = 7;
  
  i = 0;
  for (;;) {
    _GetBits("MAIN", i++ % 80);
  }
}

/*********************************************************************
*
*       main()
*
*  Function description
*   Low-level application entry point.
*   Prepare the system to run on the PSP and start it.
*/
__attribute__((naked)) int main(void) {
  asm("push {r0, lr}\n"
      "ldr r0, =__stack_process_end__\n"  // Initialize PSP
      "msr psp, r0\n"
      "mrs r0, control\n"                 // Switch to PSP
      "orr r0, #0x2\n"
      "msr control, r0\n"
      "bl main_app\n"                     // Call application main
      "pop {r0, pc}\n"
      );
}

/*************************** End of file ****************************/
