/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : main_Recursion.c
Purpose : Example application for Stack Overflow Prevention.

Literature:
  [1]  https://wiki.segger.com/Stack_Overflow_Prevention

Additional information:
  This application demonstrates how a stack overflow can happen when
  recursion is used and the worst case call graph is not correctly
  tested or identified to calculate the stack requirements.
  Depending on the input a system might usually run, but requires
  more stack and fails with different input.
  In this example sorting an unsorted array succeeds,
  sorting the sorted array again fails.
  
  Expected output:
    #### Sorting:
    32, 84, 36, 88, 90, 25, 78, 34, 61, 87, 53, 49, 65, 80, 90, 68, 
    28, 67, 79, 24,  7, 90, 27, 58, 73, 55, 54, 48, 88, 26, 33,  5, 
    93,  2, 63, 93, 21, 78, 70, 90,  1, 30, 54, 97, 46, 52, 18, 21, 
    47,  8, 92, 93, 72, 36, 22, 30, 89, 98, 28, 42, 22, 53, 69, 30, 
    73, 53, 64, 85, 76, 73, 85, 86, 66, 38, 10, 50, 13, 55, 80, 64, 
    78, 35, 13, 12, 59, 13, 15, 31,  9, 91, 38, 93,  7, 52, 90, 32, 
    26, 86, 32, 95, 28, 16, 30,  8, 12, 44, 99, 32, 47, 65, 52, 56, 
    78,  2, 69, 26,  8, 22, 24, 65, 11,  6, 32, 15, 62, 17, 15, 35
    Lvl  0: Slice: [  0:127] Pivot: 35
    Lvl  1: Slice: [  0: 51] Pivot: 15
    Lvl  2: Slice: [  0: 19] Pivot: 15
    ...
    Lvl  9: Slice: [ 44: 47] Pivot: 32
    Lvl 10: Slice: [ 44: 46] Pivot: 32
    Lvl 11: Slice: [ 44: 45] Pivot: 32
    Lvl  1: Slice: [ 53:127] Pivot: 72
    Lvl  2: Slice: [ 53: 91] Pivot: 62
    ...
    Lvl  7: Slice: [120:121] Pivot: 93
    Lvl  4: Slice: [125:127] Pivot: 97
    Lvl  5: Slice: [126:127] Pivot: 98
    Sorted.

    #### Sorting:
     1,  2,  2,  5,  6,  7,  7,  8,  8,  8,  9, 10, 11, 12, 12, 13, 
    13, 13, 15, 15, 15, 16, 17, 18, 21, 21, 22, 22, 22, 24, 24, 25, 
    26, 26, 26, 27, 28, 28, 28, 30, 30, 30, 30, 31, 32, 32, 32, 32, 
    32, 33, 34, 35, 35, 36, 36, 38, 38, 42, 44, 46, 47, 47, 48, 49, 
    50, 52, 52, 52, 53, 53, 53, 54, 54, 55, 55, 56, 58, 59, 61, 62, 
    63, 64, 64, 65, 65, 65, 66, 67, 68, 69, 69, 70, 72, 73, 73, 73, 
    76, 78, 78, 78, 78, 79, 80, 80, 84, 85, 85, 86, 86, 87, 88, 88, 
    89, 90, 90, 90, 90, 90, 91, 92, 93, 93, 93, 93, 95, 97, 98, 99
    Lvl  0: Slice: [  0:127] Pivot: 99
    Lvl  1: Slice: [  0:126] Pivot: 98
    Lvl  2: Slice: [  0:125] Pivot: 97
    ...
    Lvl 42: Slice: [  0: 85] Pivot: 65
    Lvl 43: Slice: [  0: 84] Pivot: 65
    Lvl 44: Slice: [  0: 83] Pivot: 65
    Lvl 45: Slice: [  0: 82] Pivot: 64
    Lvl 46: Slice: [  0: 81] Pivot: 64
    SYSTEM ABORT: Stack Overflow Prevented at 0x080008AE.
                  SP:      0x2001F840
                  Limit:   0x2001F858
*/

#include <stdio.h>
#include <stdlib.h>
#include "SEGGER_STOP.h"

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static int Arr[] = { 
  32, 84, 36, 88, 90, 25, 78, 34, 61, 87, 53, 49, 65, 80, 90, 68, 
  28, 67, 79, 24,  7, 90, 27, 58, 73, 55, 54, 48, 88, 26, 33,  5, 
  93,  2, 63, 93, 21, 78, 70, 90,  1, 30, 54, 97, 46, 52, 18, 21, 
  47,  8, 92, 93, 72, 36, 22, 30, 89, 98, 28, 42, 22, 53, 69, 30, 
  73, 53, 64, 85, 76, 73, 85, 86, 66, 38, 10, 50, 13, 55, 80, 64, 
  78, 35, 13, 12, 59, 13, 15, 31,  9, 91, 38, 93,  7, 52, 90, 32, 
  26, 86, 32, 95, 28, 16, 30,  8, 12, 44, 99, 32, 47, 65, 52, 56, 
  78,  2, 69, 26,  8, 22, 24, 65, 11,  6, 32, 15, 62, 17, 15, 35
};

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/
static void Print(int* pArr, int Len, int Cols) {
  if (Len > 0) {
    for (int i = 0; i < Len - 1; i++) {
      if ((i % Cols) == 0) {
        printf("\n");
      }
      printf("%2d, ", pArr[i]);
    }
    printf("%2d\n", pArr[Len - 1]);
  }
}

static void PrintProcess(int Level, int Low, int High, int Pivot) {
  printf("Lvl %2d: Slice: [%3d:%3d] Pivot: %d\n", Level, Low, High, Pivot);
}

static void Swap(int* pArr, int Left, int Right) {
  int Tmp;
  
  Tmp = pArr[Left];
  pArr[Left] = pArr[Right];
  pArr[Right] = Tmp;
}

static int IsSorted(int* Arr, int NumItems) {
  for (int n = 1; n < NumItems; n++) {
    if (Arr[n -1] > Arr[n]) {
      return 0;
    }
  }
  return 1;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
#ifdef _SORT_FAST
/*********************************************************************
*
*       QuickSort()
*
*  Function description
*   Hoare's partition scheme.
*
*  Parameters
*   pArr:   Pointer to complete array to be sorted.
*   Low:    Lowest index of range to sort (inclusive).
*   High:   Highest index of range to sort (inclusive).
*
*  Return value
*   Pivot point of range.
*/
int Partition(int* pArr, int Low, int High);
int Partition(int* pArr, int Low, int High) {
  int PivotPoint;
  int Pivot;
  int Left;
  int Right;
  
  PivotPoint = ((High - Low) / 2) + Low;
  Pivot = pArr[PivotPoint];
  Left = Low - 1;
  Right = High + 1;
  
  for (;;) {
    do {
      Left++;
    } while (pArr[Left] < Pivot);
    do {
      Right--;
    } while (pArr[Right] > Pivot);
    if (Left >= Right) {
      break;
    }
    Swap(pArr, Left, Right);
  }
  
  return Right; 
}

/*********************************************************************
*
*       QuickSort()
*
*  Function description
*   Implementation of QuickSort, using Hoare's partition scheme.
*
*  Parameters
*   Level:  Current recursion level.
*   pArr:   Pointer to complete array to be sorted.
*   Low:    Lowest index of range to sort (inclusive).
*   High:   Highest index of range to sort (inclusive).
*/
void QuickSort(int Level, int* pArr, int Low, int High);
void QuickSort(int Level, int* pArr, int Low, int High) {
  int PivotPoint;
  
  if ((Low >= 0) && (High >= 0) && (Low < High)) {
    PivotPoint = Partition(pArr, Low, High);
    PrintProcess(Level, Low, High, Arr[PivotPoint]);
    QuickSort(Level + 1, pArr, Low, PivotPoint);
    QuickSort(Level + 1, pArr, PivotPoint + 1, High);
  }
}
#else

/*********************************************************************
*
*       QuickSort()
*
*  Function description
*   Implementation of QuickSort, using the last item as pivot point.
*
*  Parameters
*   Level:  Current recursion level.
*   Arr:    Pointer to complete array to be sorted.
*   Low:    Lowest index of range to sort (inclusive).
*   High:   Highest index of range to sort (inclusive).
*/
void QuickSort(int Level, int* Arr, int Low, int High);
void QuickSort(int Level, int* Arr, int Low, int High) {
  int PivotPoint;
  int Left;
  int Right;
  
  if ((Low >= 0) && (High >= 0) && (Low < High)) {
    PivotPoint = High;
    PrintProcess(Level, Low, High, Arr[PivotPoint]);
    Left = Low - 1;
    for (Right = Low; Right <= High - 1; Right++) {
      if (Arr[Right] <= Arr[PivotPoint]) {
        Left++;
        Swap(Arr, Left, Right);
      }
    }
    Swap(Arr, Left + 1, PivotPoint);
    PivotPoint = Left + 1;
    QuickSort(Level + 1, Arr, Low, PivotPoint - 1);
    QuickSort(Level + 1, Arr, PivotPoint + 1, High);
  }
}
#endif

/*********************************************************************
*
*       main()
*
*  Function description
*   Application entry point.
*/
int main(void) {
  int NumItems;
  
#if !defined (__SEGGER_LINKER)
  //
  // Optionally re-initialize stack limits if not done by runtime init.
  //
  __SEGGER_STOP_X_InitLimits();
#endif
  
  NumItems = (sizeof(Arr) / sizeof(Arr[0]));
  printf("#### Sorting:");
  Print(Arr, NumItems, 16);
  QuickSort(0, Arr, 0, NumItems - 1);
  if (IsSorted(Arr, NumItems) == 0) {
    printf("ERROR: Failed to sort array.\n");
  } else {
    printf("Sorted.\n\n");
  }
  printf("#### Sorting:");
  Print(Arr, NumItems, 16);
  QuickSort(0, Arr, 0, NumItems - 1);
  if (IsSorted(Arr, NumItems) == 0) {
    printf("ERROR: Failed to sort array.\n");
  } else {
    printf("Sorted.\n\n");
  }

  for (;;) { ; }
}

/*************************** End of file ****************************/
