/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : main_SimpleOS.c
Purpose : Example application for Stack Overflow Prevention.

Literature:
  [1]  https://wiki.segger.com/Stack_Overflow_Prevention

Additional information:
  This application demonstrates stack overflow prevention and error 
  handling in a "multi tasking" system.
  It implementes a simplified operating system with two tasks.
  
  Each task runs on its own stack and uses the PSP.
  The task switching routine (SysTick_Handler and PendSV_Handler)
  needs to make sure to adjust __SEGGER_STOP_Limit_PSP before
  switching to another task.
  
  When one task runs out of stack, it is kept in the endless loop
  of the stack check error handler,
  while other tasks continue running as normal.
  
  Expected output
    Main Task running. (0)
    BG Task running. (0)
    BG Task running. (1)
    BG Task running. (2)
    BG Task running. (3)
    BG Task running. (4)
    Main Task running. (1)
    BG Task running. (5)
    BG Task running. (6)
    BG Task running. (7)
    BG Task running. (8)
    BG Task running. (9)
    Main Task running. (2)
    BG Task running. (10)
    Main Task running. (3)
    Main Task running. (4)
    Main Task running. (5)
    Main Task running. (6)
    ...
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#define SYSTICK_CTRL (*(volatile unsigned int*)0xE000E010)
#define SYSTICK_LOAD (*(volatile unsigned int*)0xE000E014)
#define SYSTICK_VAL  (*(volatile unsigned int*)0xE000E018)

#define _PRESERVED_REGS_SIZE (4u * 8) // R4 - R11
#define _INT_FRAME_SIZE      (4u * 8) // xPSR, ReturnAddress, LR(R14), R12, R3, R2, R1, and R0.

/*********************************************************************
*
*       Types, local
*
**********************************************************************
*/
typedef void(*TaskFunc)(void);

typedef struct Task_s Task;
struct Task_s {
  struct Task_s* pNext;
  unsigned char* pStack;
  unsigned char* pStackBottom;
  unsigned char* pStackTop;
  TaskFunc pFunc;
};

struct OS_s {
  Task*                 pCurTask;
  Task*                 pFirstTask;
  Task*                 pLastTask;
  volatile unsigned int Tick;
} OS;


/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static Task         _MainTask;
static unsigned int _MainStack[1024/4];

static Task         _BGTask;
static unsigned int _BGStack[512/4];

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/
static void _Delay(unsigned int Ticks) {
  Ticks += OS.Tick;
  while (Ticks > OS.Tick) {
    ;
  }
}

static unsigned int _Factorial(unsigned int F) {
  if (F == 1) {
    return 1;
  }
  return F * _Factorial(F -1);
}

static void _Main(void) {
  int i = 0;
  for (;;) {
    printf("Main Task running. (%d)\n", i++);
    _Delay(100);
  }
}

static void _BG(void) {
  int i = 0;
  for (;;) {
    printf("BG Task running. (%d)\n", i++);
    _Delay(20);
    _Factorial(i);
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*   SysTick interrupt handler.
*   Increment system tick and perform task switch.
*/
__attribute__((naked)) void SysTick_Handler(void) {
  asm("nop\n"
      "ldr      r1, =OS\n"
      "ldr      r0, [r1, #12]\n"  // Tick = OS.Tick
      "adds     r0, #1\n"         // Tick++
      "str      r0, [r1, #12]\n"  // OS.Tick = Tick
      "ldr      r0, [r1, #0]\n"   // OS.pCurTask
      "mrs      r3, psp\n"        // Get current task stack
      "sub      r3, #32\n"        // Preserve space on stack
      "stmia    r3, {r4-r11}\n"   // Save additional registers
      "str      r3, [r0, #4]\n"   // OS.pCurTask->pStack = $PSP
      "ldr      r3, [r0, #0]\n"   // OS.pCurTask->pNext
      "str      r3, [r1, #0]\n"   // OS.pCurTask = OS.pCurTask->pNext
      "ldr      r0, [r1, #0]\n"   // OS.pCurTask
      "ldr      r3, [r0, #8]\n"   // OS.pCurTask->pStackBottom
      "add      r3, #100\n"       // Buffer before stack overflow
      "ldr      r2, =__SEGGER_STOP_Limit_PSP\n"
      "str      r3, [r2]\n"       // Update stack limit
      "ldr      r3, [r0, #4]\n"   // OS.pCurTask->pStack
      "ldmia    r3, {r4-r11}\n"   // Restore additional registers
      "add      r3, #32\n"        // Restore preserved space
      "msr      psp, r3\n"        // Switch stack
      "bx       lr\n"             // Return to task
      );
}

/*********************************************************************
*
*       PendSV_Handler()
*
*  Function description
*   PendSV Interrupt handler used to initialize the tasking system.
*/
__attribute__((naked)) void PendSV_Handler(void) {
  asm("nop\n"
      "orr lr, lr, #0xD\n"        // Change exception LR to switch to process stack on exception return
      "ldr      r3, =OS\n"
      "add      r1, r3, #4\n"     // OS.pFirstTask
      "ldr      r2, [r1]\n"
      "add      r1, r3, #0\n"     // OS.pCurTask
      "str      r2, [r1]\n"
      "ldr      r0, [r2,#4]\n"    // OS.pFirstTask->pStack
      "add      r0, #0x20\n"      // Skip preserved regs
      "ldr      r3, [r2, #8]\n"   // OS.pCurTask->pStackBottom
      "add      r3, #100\n"       // Buffer before stack overflow
      "ldr      r2, =__SEGGER_STOP_Limit_PSP\n"
      "str      r3, [r2]\n"       // Update stack limit
      "mrs      r3, control\n"    // Switch to PSP
      "orr      r3, #0x2\n"
      "msr      control, r3\n"
      "msr      psp, r0\n"        // Set task stack
      "bx       lr\n"             // Return from interrupt into task
  );
}

/*********************************************************************
*
*       OS_InitTask()
*
*  Function description
*   Initialize a task context and add it to the list.
*/
void OS_InitTask(Task* pTask, unsigned char* pStack, unsigned int StackSize, TaskFunc pFunc);
void OS_InitTask(Task* pTask, unsigned char* pStack, unsigned int StackSize, TaskFunc pFunc) {
  unsigned int* pFrame;
  //
  // Initialize context
  //
  pTask->pFunc        = pFunc;
  pTask->pStackBottom = pStack;
  pTask->pStackTop    = pStack + StackSize - 8;
  pTask->pStack       = pTask->pStackTop - (_INT_FRAME_SIZE + _PRESERVED_REGS_SIZE);
  memset(pTask->pStack, 0, _INT_FRAME_SIZE + _PRESERVED_REGS_SIZE);
  //
  // Set artificial interrupt stack frame
  //
  pFrame = (unsigned int*)(pTask->pStack + _PRESERVED_REGS_SIZE);
  pFrame[0] = 0;  // R0
  pFrame[1] = 1;  // R1
  pFrame[2] = 2;  // R2
  pFrame[3] = 3;  // R3
  pFrame[4] = 12; // R12
  pFrame[5] = (unsigned int)pFunc;  // LR
  pFrame[6] = (unsigned int)pFunc;  // ReturnAddress
  pFrame[7] = 0x01000000;  // XPSR
  //
  // Add task to list
  //
  if (OS.pLastTask != NULL) {
    OS.pLastTask->pNext = pTask;
    pTask->pNext = OS.pFirstTask;
  } else {
    pTask->pNext = pTask;
  }
  OS.pLastTask = pTask;
  if (OS.pFirstTask == NULL) {
    OS.pFirstTask = pTask;
  }
}

/*********************************************************************
*
*       main()
*
*  Function description
*   Application entry point.
*/
int main(void) {
  asm("cpsid   i"); // Disable interrupts
  //
  // Create tasks
  //
  OS_InitTask(&_MainTask, (unsigned char*)_MainStack, sizeof(_MainStack), _Main);
  OS_InitTask(&_BGTask,   (unsigned char*)_BGStack,   sizeof(_BGStack),   _BG);
  //
  // Enable SysTick timer interrupt
  //
  SYSTICK_LOAD  = 10000;
  SYSTICK_VAL   = 0;
  SYSTICK_CTRL  = 7;
  //
  // Start the OS via PendSV
  //
  asm("ldr r1, =0xE000ED04\n"     // NVIC_INT_CTRL
      "ldr r2, =0x10000000\n"     // NVIC_PENDSVSER
      "str r2, [r1]\n"
      "dsb\n"
      "isb\n"
      "cpsie i\n"                 // Enable interrupts
      :             // Output Operands
      :             // Input Operands
      : "r1", "r2"  // Clobbers
      );
  for (;;) { ; }
}

/*************************** End of file ****************************/
