/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

Purpose : Example implementation for Stack Overflow Prevention handlers.

Literature:
  [1]  https://wiki.segger.com/Stack_Overflow_Prevention
  
Additional information:
  Stack Overflow Prevention Check is done in SEGGER_STOP_*.s,
  which calls user-provided function __SEGGER_STOP_X_OnError().
  
  The stack limit variables
  __SEGGER_STOP_Limit_MSP and __SEGGER_STOP_Limit_PSP are defined 
  in SEGGER_STOP_*.s and by default initialized by runtime init to 
  the lower bounds of the stacks (plus buffer).
  
  Optionally __SEGGER_STOP_X_InitLimits() can be called to adjust
  the defaults, e.g. to disable checks on process stack if it is unused.
  With SEGGER Linker:
    Add to the linker script:
    initialize by calling __SEGGER_STOP_X_InitLimits   { section .data.stop.* };
  With GNU Linker:
    Call __SEGGER_STOP_X_InitLimits() at the beginning of main.
    
  This example can be used as a starting point to implement 
  Stack Overflow Prevention in a system.
  It includes some configuration to demonstrate some applications.
  
  Configuration
    __SEGGER_STOP_RESET_STACK
      If == 3 (default)
        Reset the overflowing stack to its initial value.
        This does not reset the system, but enables error handlers to
        run within a valid stack area.
        Requires a global process stack.
        Requires __stack_process_end__ to be defined (by linker)).
      If == 2
        Reset the process stack to its initial value.
        Requires a global process stack.
        Requires __stack_process_end__ to be defined (by linker).
      If == 1
        Reset the main stack to its initial value.
        This does not reset the system, but enables error handlers to
        run within a valid stack area.
        Requires __stack_end__ to be defined (by linker).
      If == 0
        Do not modify stacks.
        Error handlers should not use any stack other than 
        for resetting the system.
    __SEGGER_STOP_ENABLE_INT
      If != 0 (default)
        Re-enable interrupts in error handler,
        if the interrupt (main) stack did not overflow.
        Interrupts can still be serviced, while the main system is halted.
      If == 0
        Keep interrupts disabled.
    __SEGGER_STOP_OVERRIDE_INIT
      If != 0 (default)
        Use __SEGGER_STOP_X_InitLimits() from this source.
      If == 0
        Use default __SEGGER_STOP_X_InitLimits() from library.
    __SEGGER_STOP_INIT_PSP
      If != 0 (default, when __SEGGER_STOP_OVERRIDE_INIT != 0)
        Initialize __SEGGER_STOP_Limit_PSP in __SEGGER_STOP_X_InitLimits().
        Should be enabled when using a global process stack.
        Requires __stack_process_start__ to be defined (by linker).
      If == 0
        Do not initialize __SEGGER_STOP_Limit_PSP.
        May be disabled when using a per-task stack
        and limit is initialized by RTOS.
*/

#include <stdio.h>
#include "SEGGER_STOP.h"

/*********************************************************************
*
*       Defines, defaults
*
**********************************************************************
*/
#ifndef   __SEGGER_STOP_RESET_STACK
  #define __SEGGER_STOP_RESET_STACK     3         // 0: Do not reset stacks. 1: Reset main stack. 2: Reset process stack. 3: Reset both stacks.
#endif
#ifndef   __SEGGER_STOP_ENABLE_INT
  #define __SEGGER_STOP_ENABLE_INT      1
#endif
#ifndef   __SEGGER_STOP_OVERRIDE_INIT
  #define __SEGGER_STOP_OVERRIDE_INIT   1
#endif
#ifndef   __SEGGER_STOP_INIT_PSP
  #define __SEGGER_STOP_INIT_PSP        1
#endif

extern unsigned char __STACKSIZE__[];             // Linker-generated symbol for system stack size.
extern unsigned char __stack_start__[];           // Linker-generated symbol for lower bound of system stack.
extern unsigned char __stack_end__[];             // Linker-generated symbol for lower bound of system stack.

extern unsigned char __STACKSIZE_PROCESS__[];     // Linker-generated symbol for process stack size.
extern unsigned char __stack_process_start__[];   // Linker-generated symbol for lower bound of process stack.
extern unsigned char __stack_process_end__[];     // Linker-generated symbol for upper bound of process stack.

/*********************************************************************
*
*       _HandleStackError()
*
*  Function description
*   Example error handler for stack overflow prevention.
*   Log error and halt system.
*
*  Parameters
*   SP:        New (overflown) stack pointer value.
*   Limit:     Stack pointer limit value.
*   Caller:    Caller of the stack checking function.
*   UsedStack: < 0: Process Stack. >= 0: Main Stack.
*
*  Additional information
*   When __SEGGER_STOP_RESET_STACK is set, the error handler runs
*   on a freshly initialized stack and may call other functions.
*   When __SEGGER_STOP_RESET_STACK is not set,
*   no other function must be called.
*
*   The error handler must not return.
*   It should trigger a system reset or stay in an endless loop.
*
*   This function should not be compiled with stack overflow checks.
*/
__attribute__((used, no_stack_overflow_check)) static void _HandleStackError(unsigned int SP, unsigned int Limit, unsigned int Caller, int UsedStack) {
#if __SEGGER_STOP_RESET_STACK
  //
  // Only call any function if the stack has been reset
  //
  fprintf(stderr, "SYSTEM ABORT: Stack Overflow Prevented at 0x%.8X.\n"
                  "              SP:      0x%.8X\n"
                  "              Limit:   0x%.8X\n",
                  Caller, SP, Limit);
#else
  (void)SP;
  (void)Limit;
  (void)Caller;
#endif
#if __SEGGER_STOP_ENABLE_INT
  //
  // Enable interrupt if main stack did not overflow.
  //
  if (UsedStack < 0) {
    asm("cpsie i");
  }
#endif
  //
  // Stay in endless loop.
  //
  for (;;) { ; }
}

/*********************************************************************
*
*       __SEGGER_STOP_X_OnError()
*
*  Function description
*   Callback called by stack check on stack overflow.
*   Reset stack pointer to safe value and call error handler.
*
*  Additional information
*   This callback is tail-called by the stack checking functions.
*   It assumes that:
*     - R3 contains the stack limit value.
*     - R12 contains the new stack pointer value.
*     - LR contains the caller of the stack checking function.
*     - The stack pointer has not been adjusted.
*     - __stack_end__ is the start of the main stack.
*     - __stack__process_end__ is the start of the process stack.
*
*   This function does not follow the regular calling convention.
*   This callback is implemented as naked to make sure the compiler
*   does not add a prologue which might use the stack.
*
*   This function must not be compiled with stack overflow checks.
*/
void __SEGGER_STOP_X_OnError(void) {
  asm(
      "cpsid i\n"                       // Disable interrupts
      "mov     r0, r12\n"               // Save overflowed SP
      "mov     r1, r3\n"                // Save SP limit
      "sub     r2, lr, #5\n"            // Save caller
      "mrs     r3, CONTROL\n"           // Get currently used stack
      "lsls    r3, #30\n"
#if __SEGGER_STOP_RESET_STACK == 3
      "ittee   pl\n"                    // Reset this stack
      "ldrpl   r12, =__stack_end__\n"
      "msrpl   msp, r12\n"
      "ldrmi   r12, =__stack_process_end__\n"
      "msrmi   psp, r12\n"
#elif __SEGGER_STOP_RESET_STACK == 2
      "itt     mi\n"                    // Reset process stack
      "ldrmi   r12, =__stack_process_end__\n"
      "msrmi   psp, r12\n"
#elif __SEGGER_STOP_RESET_STACK == 1
      "itt     pl\n"                    // Reset main stack
      "ldrpl   r12, =__stack_end__\n"
      "msrpl   msp, r12\n"
#endif
      "bl _HandleStackError\n"            // Call error handler
      "b .\n"                             // Stay here
      );
}

/*********************************************************************
*
*       __SEGGER_STOP_X_InitLimits()
*
*  Function description
*   Initialize limit symbols for stack overflow prevention.
*/
#if __SEGGER_STOP_OVERRIDE_INIT
void __SEGGER_STOP_X_InitLimits(void) {
  __SEGGER_STOP_Limit_MSP = (int)__stack_start__ + __SEGGER_STOP_LIMIT_BUFFER;
#if __SEGGER_STOP_INIT_PSP
  //
  // If no process stack is defined, disable checks on it.
  //
  if ((int)__STACKSIZE_PROCESS__ < __SEGGER_STOP_LIMIT_BUFFER) {
    __SEGGER_STOP_Limit_PSP = 0;
  } else {
    __SEGGER_STOP_Limit_PSP = (int)__stack_process_start__ + __SEGGER_STOP_LIMIT_BUFFER;
  }
#endif
}
#endif

/*************************** End of file ****************************/