/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

Purpose : Implementation for Stack Overflow Prevention checks.

Literature:
  [1]  https://wiki.segger.com/Stack_Overflow_Prevention
*/

#include "SEGGER_STOP.h"

        .syntax unified

#ifndef   __SEGGER_STOP_INIT_PSP
  #define __SEGGER_STOP_INIT_PSP 0
#endif

/*********************************************************************
*
*       Macros
*
**********************************************************************
*/
//
// Declare a regular function.
// Functions from the startup are placed in the init section.
//
.macro START_FUNC Name
        .section .text.\Name, "ax"
        .global \Name
        .balign 2
        .thumb_func
        .code 16
\Name:
.endm

//
// Declare a weak function
//
.macro WEAK_FUNC Name
        .section .text.\Name, "ax", %progbits
        .weak \Name
        .balign 2
        .thumb_func
        .code 16
\Name:
.endm

//
// Mark the end of a function and calculate its size
//
.macro END_FUNC name
        .size \name,.-\name
.endm

/*********************************************************************
*
*       Global data
*
**********************************************************************
*/
        .global __SEGGER_STOP_Limit_MSP
        .type __SEGGER_STOP_Limit_MSP, %object

.section .data.stop.__SEGGER_STOP_Limit_MSP, "aw"
        .balign 4
__SEGGER_STOP_Limit_MSP:
        .word __stack_start__ + __SEGGER_STOP_LIMIT_BUFFER
        .size __SEGGER_STOP_Limit_MSP,.-__SEGGER_STOP_Limit_MSP

#if __SEGGER_STOP_USE_PSP
        .global __SEGGER_STOP_Limit_PSP
        .type __SEGGER_STOP_Limit_PSP, %object
.section .data.stop.__SEGGER_STOP_Limit_PSP, "aw"
        .balign 4
__SEGGER_STOP_Limit_PSP:
#if __SEGGER_STOP_INIT_PSP
        .word __stack_process_start__ + __SEGGER_STOP_LIMIT_BUFFER
#else
        .word 0x00000000
#endif
        .size __SEGGER_STOP_Limit_PSP,.-__SEGGER_STOP_Limit_PSP
#endif

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/
/*********************************************************************
*
*       __SEGGER_STOP_GROW_R3
*
*  Function description
*    Check stack and adjust SP if there is enough space.
*
*  Additional information
*    Required additional space on stack is passed in R3.
*
*    This function does not follow the regular calling convention.
*    It must not modify any register other than SP, R12, and R3.
*    It must not use the stack.
*
*    On stack overflow this function must not return.
*    On stack overflow it jumps to __SEGGER_STOP_ON_ERROR.
*/
#undef L
#define L(label) .L_GROW_R3_##label

START_FUNC __SEGGER_STOP_GROW_R3
#if __SEGGER_STOP_USE_PSP
        //
        // Check which stack pointer is currently used:
        // bit 1 of CONTROL register 0 -> MSP, 1 -> PSP
        //
        mrs     r12, CONTROL
        lsls    r12, #30
        //
        // Calculate the new stack pointer position
        //
        sub     r12, sp, r3
        //
        // Load corresponding stack limit
        //
        ite     mi
        ldrmi   r3, =__SEGGER_STOP_Limit_PSP
        ldrpl   r3, =__SEGGER_STOP_Limit_MSP
        ldr     r3, [r3]
#else
        //
        // Calculate the new stack pointer position
        //
        sub     r12, sp, r3
        //
        // Load stack limit
        //
        ldr     r3, =__SEGGER_STOP_Limit_MSP
        ldr     r3, [r3]
#endif
        //
        // Compare with new stack value
        //
        cmp     r12, r3
        blt     L(MightError)
L(Done):
        //
        // Check passed.
        // Adjust stack and return
        //
        mov     sp, r12
        bx      lr
L(MightError):
        //
        // A stack limit value of 0 means: Check is disabled
        //
        cmp     r3, #0
        beq     L(Done)
        b       __SEGGER_STOP_X_OnError
END_FUNC __SEGGER_STOP_GROW_R3

/*********************************************************************
*
*       __SEGGER_STOP_GROW_R4
*
*  Function description
*    Check stack and adjust SP if there is enough space.
*
*  Additional information
*    Required additional space on stack is passed in R4.
*
*    This function does not follow the regular calling convention.
*    It must not modify any register other than SP, R12, and R4.
*    It must not use the stack.
*
*    On stack overflow this function must not return.
*    On stack overflow it jumps to __SEGGER_STOP_ON_ERROR.
*/
#undef L
#define L(label) .L_GROW_R4_##label

START_FUNC __SEGGER_STOP_GROW_R4
#if __SEGGER_STOP_USE_PSP
        //
        // Check which stack pointer is currently used:
        // bit 1 of CONTROL register 0 -> MSP, 1 -> PSP
        //
        mrs     r12, CONTROL
        lsls    r12, #30
        //
        // Calculate the new stack pointer position
        //
        sub     r12, sp, r4
        //
        // Load corresponding stack limit
        //
        ite     mi
        ldrmi   r4, =__SEGGER_STOP_Limit_PSP
        ldrpl   r4, =__SEGGER_STOP_Limit_MSP
        ldr     r4, [r4]
#else
        //
        // Calculate the new stack pointer position
        //
        sub     r12, sp, r4
        //
        // Load stack limit
        //
        ldr     r4, =__SEGGER_STOP_Limit_MSP
        ldr     r4, [r4]
#endif
        //
        // Compare with new stack value
        //
        cmp     r12, r4
        blt     L(MightError)
L(Done):
        //
        // Check passed.
        // Adjust stack and return
        //
        mov     sp, r12
        bx      lr
L(MightError):
        //
        // A stack limit value of 0 means: Check is disabled
        //
        cmp     r4, #0
        beq     L(Done)
        //
        // Store limit value in R3
        //
        mov     r3, r4
        b       __SEGGER_STOP_X_OnError
END_FUNC __SEGGER_STOP_GROW_R4

/*********************************************************************
*
*       __SEGGER_STOP_GROW_0
*
*  Function description
*    Check stack after SP has been adjusted.
*
*  Additional information
*    Required additional space has already been allocated on stack.
*
*    This function does not follow the regular calling convention.
*    It must not modify any register other than SP, R12.
*    It must not use the stack.
*
*    On stack overflow this function must not return.
*    On stack overflow it jumps to __SEGGER_STOP_ON_ERROR.
*/
#undef L
#define L(label) .L_GROW_0_##label

START_FUNC __SEGGER_STOP_GROW_0
#if __SEGGER_STOP_USE_PSP
        //
        // Check which stack pointer is currently used:
        // bit 1 of CONTROL register 0 -> MSP, 1 -> PSP
        //
        mrs     r12, CONTROL
        lsls    r12, #30
        //
        // Load corresponding stack limit
        //
        ite     mi
        ldrmi   r12, =__SEGGER_STOP_Limit_PSP
        ldrpl   r12, =__SEGGER_STOP_Limit_MSP
        ldr     r12, [r12]
#else
        //
        // Load stack limit
        //
        ldr     r12, =__SEGGER_STOP_Limit_MSP
        ldr     r12, [r12]
#endif
        //
        // Compare with current stack value
        //
        cmp     sp, r12
        blt     L(MightError)
L(Done):
        //
        // Check passed. Return.
        //
        bx      lr
L(MightError):
        //
        // A stack limit value of 0 means: Check is disabled
        //
        cmp     r12, #0
        beq     L(Done)
        //
        // Store limit value in R3
        // Store overflowed SP in R12
        //
        mov     r3, r12
        mov     r12, sp
        b       __SEGGER_STOP_X_OnError
END_FUNC __SEGGER_STOP_GROW_0

/*********************************************************************
*
*       __SEGGER_STOP_X_OnError
*
*  Function description
*    Default callback called on stack overflow.
*/
#undef L
#define L(label) .L_X_ON_ERROR_##label
WEAK_FUNC __SEGGER_STOP_X_OnError
L(Loop):
        b       L(Loop)
END_FUNC __SEGGER_STOP_X_OnError


/*********************************************************************
*
*       __SEGGER_STOP_X_InitLimits
*
*  Function description
*    Default implementation to initialize limit variables
*/
#undef L
#define L(label) .L_X_InitLimits_##label
WEAK_FUNC __SEGGER_STOP_X_InitLimits
        ldr     r0, =__stack_start__
        adds    r0, #__SEGGER_STOP_LIMIT_BUFFER
        ldr     r1, =__SEGGER_STOP_Limit_MSP
        str     r0, [r1]
#if __SEGGER_STOP_USE_PSP && __SEGGER_STOP_INIT_PSP
        //
        // If no process stack is defined, disable checks on it.
        //
        ldr     r0, =__STACKSIZE_PROCESS__
        cmp     r0, #__SEGGER_STOP_LIMIT_BUFFER
        itte    gt
        ldrgt   r0, =__stack_process_start__
        addgt   r0, #__SEGGER_STOP_LIMIT_BUFFER
        movle   r0, #0
        ldr     r1, =__SEGGER_STOP_Limit_PSP
        str     r0, [r1]
#endif
        bx lr
END_FUNC __SEGGER_STOP_X_InitLimits

/**************************** End of file ***************************/
